from __future__ import annotations
import re
import os
from typing import final
import pygame
import scpup


__all__ = [
  "EauText",
  "EauTextsGroup"
]


class EauTextMeta(type):
  """Metaclass for EauText

  To be able to render text you'll need to call `set_font` and provide the path
  of the font to use before trying to render any text.

  TODO: Add a default font so it doesn't raise an error when the font is not
  specified

  Attributes:
    _fontpath_:
      The path of the font currently being used. For now it is recommended to
      only use 1 font thoughout the whole game
    _fonts_:
      A dictionary in which each key relates to a font size and the value of
      each key is the Font object of that font size
  """
  _fontpath_: str | None = None
  _fonts_: dict[int, pygame.font.Font] = {}

  def set_font(cls, *paths) -> None:
    """Set the path of the font thats going to be used

    Args:
      *path:
        The path segments of the font. The font file has to be placed anywhere
        under the `<root>/assets/fonts/` directory
    """
    fontpath = os.path.join(scpup.BASE_PATH, "fonts", *paths)
    if os.path.exists(fontpath):
      cls._fontpath_ = fontpath
      cls._fonts_ = {}
    else:
      raise ValueError(f"Path '{fontpath}' does not exist.")

  def font(cls, fontsize: int, /):
    """Get a font object (or create it if it doesn't exist) given a font size.

    Args:
      fontsize:
        The size of the font

    Raises:
      AttributeError:
        Trying to call this method before `set_font`

    Returns:
      A Font object of the requested size
    """
    if fontsize not in cls._fonts_:
      cls._fonts_[fontsize] = pygame.font.Font(cls._fontpath_, fontsize)
    return cls._fonts_[fontsize]


@final
class EauText(metaclass=EauTextMeta):
  """A text object used for rendering text in pygame surfaces.

  This class supports plain texts or updateable texts (explained below), and
  also it automatically wraps the text in new lines if needed (When specifying
  maxwidth). This class also supports subgrouping text so that you can easily
  search and find a group of texts within a EauTextsGroup.

  Remember to call `EauText.set_font` before instanciating this class.

  Attributes:
    _text:
      The actual text that this object stores
    image:
      The surface where the text is rendered
    rect:
      The rectangle of the surface where the text is rendered
    updatedict:
      A dictionary of the values that can be updated where each key is the name
      of the updateable value and each value is the current value of that
      updateable value. To set an updateable value in a text pass it in between
      brackets, for example:

      ```python
      txt = EauText("Hello {name}! How are you?", ...)
      ```

      In this example updatedict would be equal to `{"name": ""}` and then
      to update that value you would do this:

      ```python
      txt.update(name="Ivan")
      ```

    color:
      A tuple of 3 numbers from 0 to 255 for the RGB color. (I don't know if you
      can add a fourth one for alpha values). Defaults to (0, 0, 0) which is
      black.
    fontsize:
      The size of the font to be used for this text.
    showing:
      Whether the text should be drawn or not
    maxwidth:
      If set, the text will be wrapped in multiple lines (If it doesn't fit the
      maxwidth). Defaults to None.
    subgroup:
      A string holding the name of the subgroup that this text belongs to.
      Defaults to None.
    _drawn_rect:
      Private attribute that stores the last rectangle drawn to a surface
    position:
      The Position of the text in an EauPosition object
      The topleft corner of the text to be rendered
  """
  __slots__ = (
    "_text",
    "image",
    "rect",
    "updatedict",
    "color",
    "fontsize",
    "showing",
    "maxwidth",
    "subgroup",
    "_drawn_rect",
    "position"
  )

  def __init__(self,
               text: str, /, *,
               fontsize: int,
               position: scpup.EauPosition,
               color: tuple[int, int, int] | None = None,
               maxwidth: int | None = None,
               subgroup: str | None = None):
    """Initializes an EauText object

    Args:
      text:
        The text to be rendered
      fontsize:
        The size of the font to use
      position:
        The Position of the text in an EauPosition object
      color:
        The color of the text. Defaults to None.
      maxwidth:
        The maximum allowed with. Defaults to None.
      subgroup:
        The name of the subgroup that this text belongs to. Defaults to None.
    """
    self._text = text
    self.fontsize = fontsize
    self.position = position
    self.color = color or (0, 0, 0)
    self.maxwidth = maxwidth
    self.subgroup = subgroup
    self.showing = True
    results = re.findall(r'{([a-zA-Z0-9_]+)}', text)
    self.updatedict = dict.fromkeys(results, '') if results else None
    self.image: pygame.Surface | None = None
    self.rect: pygame.rect.Rect | None = None
    self._drawn_rect: pygame.Rect | None = None
    if not self.updatedict:
      self._render()

  def __eq__(self, value: object) -> bool:
    if isinstance(value, EauText):
      return self._text == value._text
    elif isinstance(value, str):
      return self._text == value
    return False

  def __str__(self) -> str:
    return self._text

  def draw(self, dest: pygame.Surface) -> None:
    """Draw the render of this text to a surface

    Args:
      dest:
        The target surface
    """
    if self.showing and self.image and self.rect:
      self._drawn_rect = dest.blit(self.image, self.rect, None)  # type: ignore

  def clear(self, dest: pygame.Surface, bg: pygame.Surface) -> None:
    """Clear the last drawn rectangle of this text from a surface using a
    background

    Args:
      dest:
        The target surface
      bg:
        The background of the target surface
    """
    if self._drawn_rect:
      dest.blit(bg, self._drawn_rect, self._drawn_rect)
      self._drawn_rect = None

  def update(self, **values) -> None:
    """Update the rendered text. Use this when an updateable value needs to be
    updated

    If this method is called without any matching keys or new values then the
    renders of the text won't get created again. Furthermore, if the text
    doesn't have values to update then the method won't do anything.

    Example:

    ```python
    # if the _text is "Values: {value1}, {value2}."
    txt.update(key1="value1", key2="value2")
    ```

    Args:
      **values:
        The dict entries to replace
    """
    if self.updatedict is None:
      return
    needs_update = False
    for key, value in values.items():
      if key in self.updatedict and self.updatedict[key] != value:
        self.updatedict[key] = value
        needs_update = True
    if needs_update:
      self._render()

  def _render(self):
    """Prepare the text to render the image surface and positionate its rect.

    This method replaces all updateable values with their current values and
    calls the method `_word_wrap` to get the surface of the text, and then gets
    the rect of the image.
    """
    text = self._text.format_map(self.updatedict) if self.updatedict else self._text
    self.image = self._word_wrap(text)
    self.rect = self.image.get_rect(**self.position.as_rectargs())

  def _word_wrap(self, text: str) -> pygame.surface.Surface:
    """Builds a surface with the text rendered.

    Args:
      text:
        The text that will be drawn on the surface

    Raises:
      ValueError:
        The word is too wide and does not fit the specified maxwidth.

    Returns:
      pygame.Surface:
        The surface with the text rendered.
    """
    font = self.__class__.font(self.fontsize)
    if self.maxwidth is None:
      return font.render(text, True, self.color, None)
    else:
      x, y, rows = 0, 0, []
      words = text.split(' ')
      space: float = font.size(' ')[0]
      maxheight = 0
      rowtext = ''
      for word in words:
        size: tuple[float, float] = font.size(word)
        if size[0] > self.maxwidth:
          raise ValueError()
        if size[0] + x <= self.maxwidth:
          rowtext += word
          x += size[0] + space
          maxheight = max(maxheight, size[1])
          if len(words) - 1 == words.index(word):
            surface = font.render(rowtext, True, self.color, None)
            rows.append((
              surface,
              (0, y)
            ))
            y += maxheight
        else:
          surface = font.render(rowtext, True, self.color, None)
          rows.append((
            surface,
            (0, y)
          ))
          rowtext = word
          y += maxheight
          x = size[0]
          maxheight = size[1]
      total_surf = pygame.Surface((self.maxwidth, maxheight))
      for row in rows:
        total_surf.blit(row[0], row[1], None)
      return total_surf

  def is_colliding(self, obj: tuple[float, float] | pygame.Rect | EauText | scpup.EauSprite) -> bool:  # type: ignore
    """Test if this text is colliding with a point, rectangle, sprite, or text

    Args:
      obj:
        The point, rectangle, sprite, or text to test collition with

    Returns:
      bool:
        Whether the text is colliding or not
    """
    if self.rect is not None:
      if isinstance(obj, (pygame.Rect, pygame.rect.Rect)):
        return self.rect.colliderect(obj)
      elif isinstance(obj, (EauText, scpup.EauSprite)):
        return bool(obj.rect) and self.rect.colliderect(obj.rect)
      else:
        return self.rect.collidepoint(obj)
    return False


@final
class EauTextsGroup:
  """A group of texts.

  This class is what EauGroup is to EauSprite. It has common methods for drawing
  texts, updating them, clearing them and detecting collitions with other texts
  and sprites

  Attributes:
    _texts:
      A list of the texts that belong to this group
  """
  __slots__ = (
    "_texts",
  )

  def __init__(self, *texts: EauText) -> None:
    """Initializes a text group

    Optionally you can pass any number of EauTexts to the constructor to add
    them to this group.
    """
    self._texts: list[EauText] = list(texts)

  def __iter__(self):
    return self._texts.__iter__()

  def __next__(self):
    return self.texts.__next__()

  def __contains__(self, value) -> bool:
    return value in self._texts

  def draw(self, dest: pygame.Surface) -> None:
    """Draw the texts to a surface

    Args:
      dest:
        The target surface
    """
    for text in self._texts:
      text.draw(dest)

  def clear(self, dest: pygame.Surface, background: pygame.Surface) -> None:
    """Clear the texts from a surface using a background

    Args:
      dest:
        The target surface
      background:
        The background of the surface
    """
    for text in self._texts:
      text.clear(dest, background)

  def empty(self) -> None:
    """Remove all texts from this group"""
    self._texts.clear()

  def add(self, text: EauText) -> None:
    """Add a text to this group

    Args:
      text:
        The text to add
    """
    self._texts.append(text)

  def remove(self, text: EauText) -> None:
    """Remove a text from this group

    Args:
      text:
        The text to remove
    """
    if text in self._texts:
      self._texts.remove(text)

  def texts(self, subgroup: str | None = None) -> list[EauText]:
    """Get all the texts of a subgroup or the whole group

    Args:
      subgroup:
        The subgroup of texts to get, if it is None then all the texts within
        this group will be retrieved. Defaults to None.

    Returns:
      list[EauText]:
        A list of texts
    """
    if subgroup is None:
      return self._texts
    else:
      return [txt for txt in self._texts if txt.subgroup == subgroup]

  def text(self, value: str, subgroup=None) -> None | EauText:
    """Get a text given its value. Optionally you can specify the subgroup so
    that the text is only searched in that subgroup.

    Args:
      value:
        The string value of the text
      subgroup:
        The name of the subrgoup to be searched. Defaults to None:

    Returns:
      The EauText object found or None if it wasn't found
    """
    return next((txt for txt in self.texts(subgroup) if txt == value), None)

  def update(self, **new_data) -> None:
    """Update the updateable texts in this group

    Args:
      **new_data:
        The values to update the texts with
    """
    for text in self._texts:
      text.update(**new_data)

  def showing(self, showing, *, subgroup=None):
    """Set the visibility of the texts in this group

    Args:
      showing:
        The value to set the texts visibility to
      subgroup:
        The subgroup to narrow the affected texts. Defaults to None.
    """
    for text in self.texts(subgroup):
      text.showing = showing

  def get_collition(self,
                    obj: tuple[float, float] | EauText | scpup.EauSprite | pygame.Rect,
                    subgroup=None) -> EauText | None:
    """Tests whether any text in this group (or subgroup if specified) collides
    an object

    Args:
      obj:
        The object to test the collition with. It can be a point, a rectangle, a
        EauSprite or another EauText
      subgroup:
        The subgroup to narrow the texts to test. Defaults to None.

    Returns:
      The EauText that the obj collided with or None if it didn't collide with
      any text
    """
    clicked_texts = [
      txt for txt in self.texts(subgroup)
      if txt.is_colliding(obj)
    ]
    return clicked_texts[0] if len(clicked_texts) > 0 else None



from __future__ import annotations
from functools import partialmethod
import scpup


__all__ = [
  "EauLayout",
  "EauPosition"
]


class EauLayout:
  __slots__ = (
    "type",
    "value",
    "offset",
    "total"
  )

  def __init__(self, type: str, value: int = 0, *, offset: int = 0, total: int = 1) -> None:
    self.type = type
    self.value = value
    self.offset = offset
    self.total = total

  @classmethod
  def _create(cls, type: str, value: int = 0, *, offset: int = 0, total: int = 1) -> "EauLayout":
    return cls(type, value, offset=offset, total=total)

  Position = partialmethod(_create, 'Position')
  Top = partialmethod(_create, 'Top')
  Bottom = partialmethod(_create, 'Bottom')
  Center = partialmethod(_create, 'Center')
  Left = partialmethod(_create, 'Left')
  Right = partialmethod(_create, 'Right')
  Grid = partialmethod(_create, 'Grid')

  def parse(self, size: int) -> int:
    if self.type == "Position":
      return self.value + self.offset
    elif self.type == "Top" or self.type == "Left":
      return abs(self.offset)
    elif self.type == "Right" or self.type == "Bottom":
      return size - abs(self.offset)
    elif self.type == "Center":
      return size // 2 + self.offset
    elif self.type == "Grid":
      if self.value > self.total:
        raise ValueError(f"Value cannot be greater than total. Value: {self.value}. Total: {self.total}")
      w = size // (self.total * 2)
      t = self.value * 2 - 1
      return w * t + self.offset
    raise ValueError(f"Unknown layout type: {self.type}")


class EauPosition:
  __slots__ = (
    "x",
    "y",
    "margin"
  )

  def __init__(self,
               x: int | EauLayout,
               y: int | EauLayout,
               *,
               margin: int | tuple[int, int] | tuple[int, int, int, int] | None = None):
    self.x = x if isinstance(x, EauLayout) else EauLayout.Position(x)
    self.y = y if isinstance(y, EauLayout) else EauLayout.Position(y)
    self.margin = margin

  def as_rectargs(self) -> dict[str, int]:
    displayService = scpup.EauService.get("EauDisplayService")
    if not displayService:
      raise RuntimeError("EauDisplayService has not been created, thus cannot convert EauPosition to rect arguments")
    size = displayService.size
    xarg = self.x.type.lower() if self.x.type in ["Left", "Right"] else "centerx"
    yarg = self.y.type.lower() if self.y.type in ["Top", "Bottom"] else "centery"
    if isinstance(self.margin, int):
      ml = mr = mt = mb = self.margin
    elif self.margin is None:
      ml = mr = mt = mb = 0
    elif len(self.margin) == 2:
      (ml, mt), (mr, mb) = self.margin, self.margin
    else:
      ml, mr, mt, mb = self.margin
    return {
      xarg: ml + self.x.parse(size[0] - (ml + mr)),
      yarg: mt + self.y.parse(size[1] - (mt + mb))
    }



from __future__ import annotations
import os
import scpup
import pygame

__all__ = [
  "load_image",
  "load_sound",
  "BASE_PATH"
]

BASE_PATH = os.path.join(os.getcwd(), "assets")


def load_image(
  *paths: str,
  alpha=True,
  position: scpup.EauPosition | None = None
) -> tuple[pygame.Surface, pygame.Rect]:
  """Loads an image located somewhere in <root>/assets/images/...

  Args:
    *paths:
      The path segments of where the image file exists.
    alpha:
      Whether the image should convert pixel format considering transparency or
      not. Defaults to True.
    position:
      The rectangle position as an EauPosition object

  Raises:
    ValueError:
      The given image path does not exist.

  Returns:
    tuple[pygame.Surface, pygame.Rect]:
      The image as a pygame.Surface and the covering rectangle of that image.
  """
  path = os.path.join(BASE_PATH, "images", *paths)
  if not os.path.exists(path):
    raise ValueError(f"Path: '{path}' does not exist")
  image = pygame.image.load(path)
  if alpha:
    image = image.convert_alpha()
  else:
    image = image.convert()
  rect = image.get_rect(**(position.as_rectargs() if position else {}))
  return image, rect


def load_sound(*paths: str) -> pygame.mixer.Sound:
  """Loads a sound file located somewhere in <root>/assets/sounds...

  Args:
    *paths:
      The path segments of where the sound file exists

  Raises:
    ValueError:
      The given sound path does not exist.

  Returns:
    pygame.mixer.Sound:
      The loaded sound.
  """
  path: str = os.path.join(BASE_PATH, "sounds", *paths)
  if not os.path.exists(path):
    raise ValueError(f"Path: '{path}' does not exist")
  sound = pygame.mixer.Sound(path)
  return sound



"""This module contains base group classes used in SCPUP

The exported classes of this module are:

* EauGroup
* EauNamedGroup
"""

from __future__ import annotations
from typing import Any, Iterator
import pygame
import scpup

__all__: list[str] = [
  "EauGroup",
  "EauNamedGroup"
]


class EauGroup:
  """A sprite group

  This sprite group class is the base clase for all the other group classes.
  This class is basically a copy of the pygame.sprite.Group class with
  __slots__ added for optimization and spritedict changed to __s, but it has a
  property called spritedict so that its compatible with pygame's groups

  Attributes:
    __s:
      Private attribute that stores the sprites that belong to this group
    lostsprites:
      Private attribute that is used to clear correctly the screen after
      removing sprites from this group
  """
  _spritegroup = True

  __slots__: tuple = (
    "__s",
    "lostsprites"
  )

  def __init__(self, *sprites: scpup.EauSprite) -> None:
    """Initializes a sprite group.

    Args:
      *sprites:
        Optionally the constructor can recieve any number of sprites that belong
        to it (To add them to this group)
    """
    self.__s: dict[scpup.EauSprite, Any] = {}
    self.lostsprites = []
    self.add(*sprites)

  def __bool__(self) -> bool:
    """Tests whether this group is empty or not

    Returns:
      bool:
        True if is not empty, False otherwise
    """
    return len(self) > 0

  def __len__(self) -> int:
    """Get the amount of sprites in this group

    Returns:
      int:
        The number of sprites that belong to this group
    """
    return len(self.__s)

  def __iter__(self) -> Iterator:
    """Returns an iterator of the sprites in this group"""
    return iter(self.sprites())

  def __contains__(self, sprite: scpup.EauSprite) -> bool:
    """Tests whether a sprite exists or not in this group

    Args:
      sprite:
        The sprite to test if belongs or not

    Returns:
      bool:
        True if the sprite belongs to this group, False otherwise
    """
    return sprite in self.__s

  def __getitem__(self, idx: int) -> scpup.EauSprite:
    """Get a sprite by index.

    I haven't tested if this method affects the performance but I think that it
    is not the best method to use.

    Args:
      idx:
        The index of the sprite.

    Raises:
      IndexError:
        Index out of range.

    Returns:
      scpup.EauSprite:
        The found sprite.
    """
    if 0 < self.__len__() < idx:
      return self.sprites()[idx]
    raise IndexError('Index out of range')

  @property
  def spritedict(self) -> dict[scpup.EauSprite, Any]:
    """Get the self.__s attribute. Basically this is just for compatibility
    issues with pygame.sprite.Group and its subclasses

    Returns:
      dict:
        The attribute __s
    """
    return self.__s

  def add_internal(self, sprite: scpup.EauSprite) -> None:
    """Private method used to register a sprite that belongs to this group

    Args:
      sprite: The sprite to register
    """
    self.__s[sprite] = 0

  def remove_internal(self, sprite: scpup.EauSprite) -> None:
    """Private method used to remove a registered sprite

    Args:
      sprite: The sprite to remove
    """
    lost_rect = self.__s[sprite]
    if lost_rect:
      self.lostsprites.append(lost_rect)
    del self.__s[sprite]

  def sprites(self) -> list[scpup.EauSprite]:
    """Get a list of the sprites in this group

    TODO: Maybethis should be a property...

    Returns:
      list:
        A list of the sprites that belong to this group
    """
    return list(self.__s)

  def add(self, *sprites: scpup.EauSprite) -> None:
    """Add sprites to this group

    Args:
      *sprites:
        The sprites to be added. These can be either pygame.srite.Sprite or
        scpup.EauSprite or any subclass of these.
    """
    for sprite in sprites:
      if sprite not in self.__s:
        self.add_internal(sprite)
        sprite.add_internal(self)

  def remove(self, *sprites: scpup.EauSprite) -> None:
    """Remove sprites from this group

    Args:
      *sprites:
        The sprites to be removed. These can be either pygame.srite.Sprite or
        scpup.EauSprite or any subclass of these
    """
    for sprite in sprites:
      if sprite in self.__s:
        self.remove_internal(sprite)
        sprite.remove_internal(self)

  def update(self, *args, **kwargs) -> None:
    """Call the update method of the sprites in this group"""
    for sprite in self.sprites():
      sprite.update(*args, **kwargs)

  def draw(self, surface: pygame.Surface) -> None:
    """Draws the sprites that belong to this group on a given surface

    Args:
      surface:
        The pygame.Surface to draw the sprites to
    """
    sprites: list[scpup.EauSprite] = self.sprites()
    self.draw_internal(surface, sprites)

  def draw_internal(self, surface: pygame.Surface, sprites: list[scpup.EauSprite]) -> None:
    """Private method used to draw the sprites to a surface

    This method is used by the draw method but exists so that subclasses can
    decide which sprites should be drawn if for any reason not all sprites need
    to be drawn

    Args:
        surface:
          The pygame.Surface to draw the sprites to
        sprites:
          The sprites to draw to the surface
    """
    if hasattr(surface, "blits"):
      self.__s.update(
        zip(
          sprites,
          surface.blits(
            (spr.image, spr.rect, None) for spr in sprites
          )  # type: ignore
        )
      )
    else:
      for spr in sprites:
        if spr.image is not None and spr.rect is not None:
          self.__s[spr] = surface.blit(spr.image, spr.rect, None)
    self.lostsprites = []

  def clear(self, surface: pygame.Surface, bg: pygame.Surface) -> None:
    """Clear the sprites that were last drawn from a surface using a background

    Args:
      surface:
        The pygame.Surface to clear the sprites from
      bg:
        The pygame.Surface that will be used to clear the sprites from the
        surface
    """
    for lost_clear_rect in self.lostsprites:
      surface.blit(bg, lost_clear_rect, lost_clear_rect)
    for clear_rect in self.__s.values():
      if clear_rect:
        surface.blit(bg, clear_rect, clear_rect)

  def empty(self) -> None:
    """Remove all sprites from this group. THIS STILL HAS THE FUNCTIONALITY IT
    HAD BEFORE COMMENTED OUT BECAUSE IT HAS NOT BEEN TESTED"""
    self.remove(*self.sprites())
    # for sprite in self.__s:
    #   self.remove_internal(sprite)
    #   sprite.remove_internal(self)


class EauNamedGroup(EauGroup):
  """A group with named subgroups for a better organization

  Attributes:
    __n:
      Private attribute that stores the subgroups names along with the sprites
      that each group contains. All named groups have an 'all' group for sprites
      that don't belong to a specific subgroup.
  """
  __slots___: tuple = (
    "__n",
  )

  def __init__(self, *sprites: scpup.EauSprite, name: str = 'all') -> None:
    """Initializes a named group. Optionally sprites can by added to a subgroup
    when initializing the group

    Args:
      *sprites:
        The sprites that will be added to the subgroup
      name:
        The name of the subgroup, by default its 'all'
    """
    super().__init__(*sprites)
    self.__n: dict[str, list[scpup.EauSprite]] = {'all': []}
    self.__n[name] = list(sprites)

  def group(self, name: str) -> list[scpup.EauSprite]:
    """Get a list of the sprites in a subgroup

    Args:
      name:
        The name of the subgroup

    Returns:
      list:
        A list of sprites found, or an empty list if the subgroup does not exist
        in this group
    """
    if name in self.__n:
      return self.__n[name]
    return []

  def add(self, *sprites: scpup.EauSprite, name: str = 'all') -> None:
    """Add sprites to a subgroup of this group

    Args:
      *sprites:
        The sprites to be added
      name:
        The name of the subgroup. 'all' by default
    """
    super().add(*sprites)
    if name != 'all':
      for s in sprites:
        self.transfer(s, name)

  def add_internal(self, sprite: scpup.EauSprite) -> None:
    super().add_internal(sprite)
    self.__n['all'].append(sprite)

  def remove_internal(self, sprite: scpup.EauSprite) -> None:
    super().remove_internal(sprite)
    sprites_list = next(filter(lambda s: sprite in s, self.__n.values()), None)
    if sprites_list:
      sprites_list.remove(sprite)

  def empty(self, name: str | None = None) -> None:
    """Remove all sprites from a subgroup of this group

    Args:
      name:
        The name of the subgroup to empty. If None is passed then empty the
        whole group
    """
    self.remove(*self.sprites(name))
    if name is None:
      self.__n.clear()
    else:
      self.__n[name] = []

  def sprites(self, name: str | None = None) -> list[scpup.EauSprite]:
    """Get a list of the sprites in a subgroup of this group

    Args:
      name:
        The name of the subgroup to search. If None is passed then get the
        sprites of the whole group

    Returns:
      list:
        A list of sprites
    """
    return super().sprites() if name is None else self.__n[name] if name in self.__n else []

  def update(self, *args, name: str | None = None, **kwargs) -> None:
    """Call the update method of the sprites in a subgroup of this group

    Example calls of this method:

    ```python
    named_group.update(name='subgroup_name')
    named_group.update('value', name='subgroup_name')
    named_group.update(name='subgroup_name', kwarg='value')
    named_group.update('value', name='subgroup_name', kwarg='value')
    ```

    Args:
      name:
        The name of the subgroup to update. If None is passed then update the
        whole group
    """
    for sprite in self.sprites(name):
      sprite.update(*args, **kwargs)

  def draw(self, surface: pygame.Surface, name: str | None = None) -> None:
    """Draws the sprites that belong to a subgroup of this group on a surface

    Args:
      name:
        The name of the subgroup to draw. If None is passed then the whole group
        is drawn
      surface:
        The pygame.Surface to draw the sprites to
    """
    sprites: list[scpup.EauSprite] = self.sprites(name)
    self.draw_internal(surface, sprites)

  def transfer(self, sprite: scpup.EauSprite | None, name: str) -> None:
    """Move a sprite from its current group to another group

    Args:
      sprite:
        The sprite to move. (Added the '| None' to the type of this parameter
        so that you can pass an object that may be a sprite or None to this
        function but if None is passed it doesn't do anything)
      name:
        The name of the group to move the sprite to. If the group does not exist
        yet it will be created.
    """
    if sprite is None:
      return
    sprites_list = next(filter(lambda s: sprite in s, self.__n.values()), None)
    if sprites_list:
      sprites_list.remove(sprite)
      if name in self.__n:
        self.__n[name].append(sprite)
      else:
        self.__n[name] = [sprite]



"""scpup package contains the classes, constants, types, etc., needed in the
game SCPUP (Super Crystal Pokebros Ultimate Party) and any other AramEau game.
"""

import pygame
from .services import EauService  # noqa
from .text import *  # noqa
from .loader import *  # noqa
from .sprite import *  # noqa
from .group import *  # noqa
from .view import *  # noqa
from .player import *  # noqa
from .ctrl import *  # noqa
from .position import *  # noqa

__name__ = "scpup"
__package__ = "scpup"


def init(window_attrs: dict):
  from .services import EauDisplayService, EauEventService
  from .text import EauText
  pygame.init()
  EauDisplayService.set_window_attrs(**window_attrs)
  EauDisplayService()
  EauEventService()
  if not EauText._fontpath_:
    raise RuntimeError('Font path has not been set!!!')



"""This module contains the helper classes for game controllers.

This module has the button and stick mappings of the following controllers:

* PS4
  * PS4 Controller
  * DualSense Wireless Controller
* PS5
  * Sony Interactive Entertainment Wireless Controller
* Xbox
  * Xbox Series X Controller
  * Xbox One Series X Controller

The mapping classes also are used to get information of each stick and trigger
"""

from __future__ import annotations
import pygame
import pygame.joystick
from enum import IntEnum, Enum, auto
from abc import abstractmethod, ABCMeta
from typing import overload, final

__all__: list[str] = [
  "EauAction",
  "EauAxis",
  "EauCtrl",
  "EauCtrlMapping",
  "EauXboxCtrlMapping",
  "EauPS4CtrlMapping",
  "EauPS5CtrlMapping",
]


@final
class EauAxis(Enum):
  """Enumeration that has all axis available. Left and Right sticks have a
  vertical and horizontal movement."""
  LS_X = auto()
  LS_Y = auto()
  RS_X = auto()
  RS_Y = auto()
  LT = auto()
  RT = auto()


@final
class EauAction(IntEnum):
  """Enumeration that has all the mappeable buttons."""
  START = auto()
  SELECT = auto()
  GUIDE = auto()
  A = auto()
  B = auto()
  X = auto()
  Y = auto()
  LB = auto()
  RB = auto()
  LS = auto()
  RS = auto()
  RT = auto()
  LT = auto()
  DPAD_UP = auto()
  DPAD_RIGHT = auto()
  DPAD_DOWN = auto()
  DPAD_LEFT = auto()
  LS_LEFT = auto()
  LS_RIGHT = auto()
  LS_UP = auto()
  LS_DOWN = auto()
  RS_LEFT = auto()
  RS_RIGHT = auto()
  RS_UP = auto()
  RS_DOWN = auto()


class EauCtrlMappingMeta(ABCMeta):
  """Metaclass of the EauCtrlMapping class. This class implements a Singleton"""
  _instances_: dict = {}

  def __call__(cls, *args, **kwargs):
    if cls not in cls._instances_:
      instance = super().__call__(*args, **kwargs)
      cls._instances_[cls] = instance
    return cls._instances_[cls]


class EauCtrlMapping(metaclass=EauCtrlMappingMeta):
  """A mapping between the buttons of a controller to the button id of a
  pygame.event.Event. This class implements the Singleton pattern."""
  __slots__: tuple = ()

  @staticmethod
  def get(name: str) -> EauXboxCtrlMapping | EauPS4CtrlMapping | EauPS5CtrlMapping:
    """Get the mapping instance of a certain controller given the name of that
    controller.

    Args:
      name:
        The name of the controller as described in the pygame.joystick
        documentation.

    Raises:
      ValueError:
        The name is not a valid controller name or it has not been mapped yet.

    Returns:
      The instance of the corresponding controller mapping.
    """
    if name in ["Xbox Series X Controller", "Xbox One Series X Controller"]:
      return EauXboxCtrlMapping()
    elif name in ["PS4 Controller", "DualSense Wireless Controller"]:
      return EauPS4CtrlMapping()
    elif name in ["Sony Interactive Entertainment Wireless Controller"]:
      return EauPS5CtrlMapping()
    raise ValueError(f"Controller not defined: '{name}'")

  @abstractmethod
  def __getitem__(self, key: int) -> EauAction:
    """Gets a scpup.EauAction member given a key (which would be the button id
    of a pygame.event.Event).

    Args:
      key:
        The button that is received from a pygame.event.Event instance when a
        JOYBUTTONUP or JOYBUTTONDOWN event is received.

    Returns:
      EauAction:
        The button that the received key maps to.

    Raises:
      KeyError:
        The key is a integer but does not correspond to a button
      TypeError:
        It could be that the key is not an integer or any other unexpected
        error.
    """

  @overload
  @abstractmethod
  def axis(self, axis: EauAxis) -> int:
    """Get the axis number given an EauAxis identifier

    Args:
      axis:
        The EauAxis identifier of the axis

    Returns:
      int:
        The axis number

    Raises:
      TypeError:
        An unknown axis was received
    """
  @overload
  @abstractmethod
  def axis(self, axis: int) -> EauAxis:
    """Get the EauAxis identifier given an axis number

    Args:
      axis:
        The axis number

    Returns:
      EauAxis:
        The EauAxis identifier corresponding to the axis number received"""
  @abstractmethod
  def axis(self, axis: EauAxis | int) -> int | EauAxis:
    ...


@final
class EauXboxCtrlMapping(EauCtrlMapping):
  __slots__: tuple = ()

  def __getitem__(self, key: int) -> EauAction:
    if key < 0 or key > 10:
      raise KeyError(f'No mapping for button {key}')
    elif key == 0:
      return EauAction.A
    elif key == 1:
      return EauAction.B
    elif key == 2:
      return EauAction.X
    elif key == 3:
      return EauAction.Y
    elif key == 4:
      return EauAction.LB
    elif key == 5:
      return EauAction.RB
    elif key == 6:
      return EauAction.SELECT
    elif key == 7:
      return EauAction.START
    elif key == 8:
      return EauAction.LS
    elif key == 9:
      return EauAction.RS
    elif key == 10:
      return EauAction.GUIDE
    raise TypeError('Unexpected Error')

  @overload
  def axis(self, axis: EauAxis) -> int: ...

  @overload
  def axis(self, axis: int) -> EauAxis: ...

  def axis(self, axis: EauAxis | int) -> int | EauAxis:
    if isinstance(axis, EauAxis):
      if axis == EauAxis.LS_X:
        return 0
      elif axis == EauAxis.LS_Y:
        return 1
      elif axis == EauAxis.RS_X:
        return 3
      elif axis == EauAxis.RS_Y:
        return 4
      elif axis == EauAxis.LT:
        return 2
      elif axis == EauAxis.RT:
        return 5
      raise TypeError(f"Unknown axis number: '{axis}'")
    else:
      if axis == 0:
        return EauAxis.LS_X
      elif axis == 1:
        return EauAxis.LS_Y
      elif axis == 3:
        return EauAxis.RS_X
      elif axis == 4:
        return EauAxis.RS_Y
      elif axis == 2:
        return EauAxis.LT
      elif axis == 5:
        return EauAxis.RT
      raise TypeError(f"Unknown axis identifier: '{axis}'")


@final
class EauPS4CtrlMapping(EauCtrlMapping):
  __slots__: tuple = ()

  def __getitem__(self, key: int) -> EauAction:
    if key < 0 or key > 14:
      raise KeyError(f'No mapping for button {key}')
    elif key == 0:
      return EauAction.A
    elif key == 1:
      return EauAction.B
    elif key == 2:
      return EauAction.X
    elif key == 3:
      return EauAction.Y
    elif key == 4:
      return EauAction.SELECT
    elif key == 5:
      return EauAction.GUIDE
    elif key == 6:
      return EauAction.START
    elif key == 7:
      return EauAction.LS
    elif key == 8:
      return EauAction.RS
    elif key == 9:
      return EauAction.LB
    elif key == 10:
      return EauAction.RB
    elif key == 11:
      return EauAction.DPAD_UP
    elif key == 12:
      return EauAction.DPAD_DOWN
    elif key == 13:
      return EauAction.DPAD_LEFT
    elif key == 14:
      return EauAction.DPAD_RIGHT
    raise TypeError('Unexpected Error')

  @overload
  def axis(self, axis: EauAxis) -> int: ...

  @overload
  def axis(self, axis: int) -> EauAxis: ...

  def axis(self, axis: EauAxis | int) -> int | EauAxis:
    if isinstance(axis, EauAxis):
      if axis == EauAxis.LS_X:
        return 0
      elif axis == EauAxis.LS_Y:
        return 1
      elif axis == EauAxis.RS_X:
        return 2
      elif axis == EauAxis.RS_Y:
        return 3
      elif axis == EauAxis.LT:
        return 4
      elif axis == EauAxis.RT:
        return 5
      raise TypeError(f"Unknown axis number: '{axis}'")
    else:
      if axis == 0:
        return EauAxis.LS_X
      elif axis == 1:
        return EauAxis.LS_Y
      elif axis == 2:
        return EauAxis.RS_X
      elif axis == 3:
        return EauAxis.RS_Y
      elif axis == 4:
        return EauAxis.LT
      elif axis == 5:
        return EauAxis.RT
      raise TypeError(f"Unknown axis identifier: '{axis}'")


@final
class EauPS5CtrlMapping(EauCtrlMapping):
  __slots__: tuple = ()

  def __getitem__(self, key: int) -> EauAction:
    if key < 0 or key > 12:
      raise KeyError(f'No mapping for button {key}')
    elif key == 0:
      return EauAction.A
    elif key == 1:
      return EauAction.B
    elif key == 2:
      return EauAction.X
    elif key == 3:
      return EauAction.Y
    elif key == 4:
      return EauAction.LB
    elif key == 5:
      return EauAction.RB
    # elif key == 6:
    #   return EauAction.LT
    # elif key == 7:
    #   return EauAction.RT
    elif key == 8:
      return EauAction.SELECT
    elif key == 9:
      return EauAction.START
    elif key == 10:
      return EauAction.GUIDE
    elif key == 11:
      return EauAction.LS
    elif key == 12:
      return EauAction.RS
    raise TypeError('Unexpected Error')

  @overload
  def axis(self, axis: EauAxis) -> int: ...

  @overload
  def axis(self, axis: int) -> EauAxis: ...

  def axis(self, axis: EauAxis | int) -> int | EauAxis:  # type: ignore
    if isinstance(axis, EauAxis):
      if axis == EauAxis.LS_X:
        return 0
      elif axis == EauAxis.LS_Y:
        return 1
      elif axis == EauAxis.RS_X:
        return 3
      elif axis == EauAxis.RS_Y:
        return 4
      elif axis == EauAxis.LT:
        return 2
      elif axis == EauAxis.RT:
        return 5
      raise TypeError(f"Unknown axis number: '{axis}'")
    else:
      if axis == 0:
        return EauAxis.LS_X
      elif axis == 1:
        return EauAxis.LS_Y
      elif axis == 3:
        return EauAxis.RS_X
      elif axis == 4:
        return EauAxis.RS_Y
      elif axis == 2:
        return EauAxis.LT
      elif axis == 5:
        return EauAxis.RT
      raise TypeError(f"Unknown axis identifier: '{axis}'")


class EauCtrlMeta(type):
  """Metaclass of the EauCtrl class. This class implements a Singleton, but
  instead of storing only 1 instance of the EauCtrl class, it holds an instance
  for every pygame.joystick.Joystick instance"""
  _instances: dict[int, EauCtrl] = {}

  @overload
  def __call__(cls, iid: int) -> EauCtrl:
    """Get an EauCtrl instance given an instance id.

    Args:
      iid:
        The instance id of the joystick wrapped by an instance of the EauCtrl
        class.

    Returns:
      EauCtrl:
        The EauCtrl instance that has the joystick with the given instance id
    """
  @overload
  def __call__(cls, joystick: pygame.joystick.Joystick) -> EauCtrl:
    """Create an instance of EauCtrl and store it (for the Singleton
    implementation).

    The Singleton implementation uses the joystick instance ids as keys to store
    EauCtrl instances

    Args:
      joystick:
        The joystick the will be wrapped by the new EauCtrl instance

    Returns:
      The new EauCtrl instance, or an existing instance if the joystick was
      already wrapped by it.
    """
  def __call__(cls, val: pygame.joystick.JoystickType | int) -> EauCtrl:  # type: ignore
    if isinstance(val, int):
      return next(ctrl for iid, ctrl in cls._instances.items() if iid == val)
    iid = val.get_instance_id()
    if iid not in cls._instances:
      instance = super().__call__(val)
      cls._instances[iid] = instance
    return cls._instances[iid]

  def remove_ctrl(cls, iid: int) -> None:
    """Remove a controller from the saved instances, given an instance id.

    Args:
      iid:
        The instance id of the joystick to look for.
    """
    if iid in cls._instances:
      del cls._instances[iid]


@final
class EauCtrl(metaclass=EauCtrlMeta):
  """A wrapper around a pygame.joystick object that also has its mapping.

  Attributes:
    joystick:
      The wrapped pygame.joystick object.
    _mapping:
      The corresponding EauCtrlMapping object.
    threshold:
      Class attribute that is used for the sticks dead zone adjustment
  """
  threshold = 0.3

  __slots__: tuple = (
    "joystick",
    "_mapping",
  )

  @overload
  def __init__(self, iid: int): ...

  @overload
  def __init__(self, joystick: pygame.joystick.Joystick): ...

  def __init__(self, joystick: pygame.joystick.Joystick) -> None:  # type: ignore
    """Initializes a controller wrapper for a given joystick.

    Args:
      joystick:
        The pygame.joystick that will be wrapped.
    """
    self._mapping: EauXboxCtrlMapping | EauPS4CtrlMapping | EauPS5CtrlMapping = EauCtrlMapping.get(joystick.get_name())
    self.joystick: pygame.joystick.Joystick = joystick

  @property
  def iid(self) -> int:
    """Gets the instance id of the wrapped joystick."""
    return self.joystick.get_instance_id()

  @property
  def LS(self) -> tuple[float, float]:
    """Get the current state of the left stick

    Returns:
      tuple[float, float]:
        Horizontal and vertical state of the left stick. Values go from -1 to 1
    """
    return (
      self.joystick.get_axis(self._mapping.axis(EauAxis.LS_X)),
      self.joystick.get_axis(self._mapping.axis(EauAxis.LS_Y))
    )

  @property
  def RS(self) -> tuple[float, float]:
    """Get the current state of the right stick

    Returns:
      tuple[float, float]:
        Horizontal and vertical state of the right stick. Values go from -1 to 1
    """
    return (
      self.joystick.get_axis(self._mapping.axis(EauAxis.RS_X)),
      self.joystick.get_axis(self._mapping.axis(EauAxis.RS_Y))
    )

  @property
  def triggers(self) -> tuple[float, float]:
    """Get the current state of the triggers

    Returns:
      tuple[float, float]:
        Left trigger and right trigger states. Values go from 0 to 1 (Not sure)
    """
    return (
      self.joystick.get_axis(self._mapping.axis(EauAxis.LT)),
      self.joystick.get_axis(self._mapping.axis(EauAxis.RT))
    )

  @overload
  def action(self, button_num: int) -> EauAction:
    """Get an action given a button number

    Args:
      button_num:
        The button number

    Returns:
      EauAction:
        The action that the button matches to.
    """
  @overload
  def action(self, axis_num: int, value: float) -> EauAction:
    """Get an action given an axis number and its value

    Args:
      axis_num:
        The axis number
      value:
        The movement of the axis

    Returns:
      EauAction:
        The action that the axis matches to. The value is used to determinate
        the direction of the movement.
    """
  def action(self, num: int, value: float | None = None) -> EauAction | None:  # type: ignore
    if value is not None:
      if -self.threshold < value < self.threshold:
        return None
      axis = self._mapping.axis(num)
      if axis == EauAxis.LS_X:
        return EauAction.LS_LEFT if value < -self.threshold else EauAction.LS_RIGHT
      elif axis == EauAxis.LS_Y:
        return EauAction.LS_UP if value < -self.threshold else EauAction.LS_DOWN
      elif axis == EauAxis.RS_X:
        return EauAction.RS_LEFT if value < -self.threshold else EauAction.RS_RIGHT
      elif axis == EauAxis.RS_Y:
        return EauAction.RS_UP if value < -self.threshold else EauAction.RS_DOWN
      elif axis == EauAxis.LT:
        return EauAction.LT
      elif axis == EauAxis.RT:
        return EauAction.RT
    else:
      try:
        action = self._mapping[num]
        return action
      except (TypeError, KeyError):
        return None



from __future__ import annotations
import scpup
import pygame
from typing import final, overload
from typing import Literal

from scpup.sprite import EauSprite

__all__: list[str] = [
  "PlayerIdType",
  "EauPlayer"
]

PlayerIdType = Literal["p1", "p2", "p3", "p4"]


class EauPlayerMeta(type):
  """Metaclass for EauPlayer class. Implements a Singleton but instead of
  storing only 1 instance, it stores a maximum of 4 instances"""
  _instances: dict[str, EauPlayer] = {}
  _current_sprite: str | None = None

  @overload
  def __call__(cls) -> list[EauPlayer]:
    """Get a list of all the player instances

    Returns:
      list[EauPlayer]:
        All the player instances
    """
  @overload
  def __call__(cls, iid: int) -> EauPlayer | None:
    """Get a player given a controller instance id.

    Args:
      iid:
        The controller instance id.

    Returns:
      EauPlayer:
        The player with the controller with the instance id assigned.
      None:
        No player has the controller with the instance id provided assigned.
    """
  @overload
  def __call__(cls, pid: str) -> EauPlayer:
    """Get a player given a player id.

    This overload will create the player if it does not exist yet.

    Args:
      pid:
        The player id. It can be one of 'p1', 'p2', 'p3', and 'p4'.

    Returns:
      EauPlayer:
        The player found or created with the given player id.
    """
  def __call__(cls, pid_iid: int | str | None) -> list[EauPlayer] | EauPlayer | None:  # type: ignore
    if pid_iid is None:
      return list(cls._instances.values())
    elif isinstance(pid_iid, int):
      return next((p for p in cls._instances.values() if p.iid == pid_iid), None)
    else:
      if pid_iid not in cls._instances:
        instance = super().__call__(pid_iid)
        cls._instances[pid_iid] = instance
      return cls._instances[pid_iid]

  def __len__(cls) -> int:
    """Get the number of player instances

    Returns:
      int:
        The number of player instances
    """
    return len(cls._instances)

  def __next__(cls) -> EauPlayer:
    """Create a player.

    This method will know which player id goes next. It was implemented this way
    so that you could do:

    ```python
    try:
      player = next(EauPlayer)
    except StopIteration:
      ...
    ```

    Raises:
      StopIteration:
        Can't create a new player because 4 players already exist.

    Returns:
      EauPlayer:
        The new player.
    """
    if len(cls._instances) < 4:
      return cls.__call__(f"p{len(cls._instances) + 1}")
    raise StopIteration

  @overload
  def remove_player(cls, iid: int) -> None:
    """Remove a player instance given a controller instance id

    Args:
      iid:
        The controller instance id to find the player that has it assigned
    """
  @overload
  def remove_player(cls, player: EauPlayer) -> None:
    """Remove a player instance.

    This method does all the logic needed to remove a player instance, like
    unassigning the controller that the player has currently assigned, and stuff
    like that.

    Args:
      player:
        The player instance to remove.
    """
  def remove_player(cls, iid_or_player: EauPlayer | int) -> None:  # type: ignore
    if isinstance(iid_or_player, int):
      player = EauPlayer(iid_or_player)
    else:
      player = iid_or_player
    if player is not None and player in cls._instances.values():
      player.unassign_ctrl()
      cls._instances.pop(player.pid)
      players = list(cls._instances.values())
      cls._instances = {}
      for i in range(len(players)):
        players[i].pid = f"p{i + 1}"
        cls._instances[players[i].pid] = players[i]

  def draw(cls, surface) -> None:
    """Calls the draw method of each EauPlayer instance

    Args:
      surface:
        The target surface
    """
    for p in cls._instances.values():
      p.sprites.draw(surface, name="active")

  def update(cls, *args, **kwargs) -> None:
    """Calls the update method of each EauPlayer instance"""
    for p in cls._instances.values():
      p.move(kwargs.get('rect', None))
      p.sprites.update(*args, name="active", **kwargs)

  def clear(cls, surface, background) -> None:
    """Calls the clear method of each EauPlayer instance

    Args:
      surface:
        The target surface
      background:
        The background surface
    """
    for p in cls._instances.values():
      p.sprites.clear(surface, background)

  def check_collitions(cls, group) -> None:
    """Check collitions for each EauPlayer instance

    Args:
      group:
        The group of sprites to test collitions against
    """
    for p in cls._instances.values():
      p.check_collition(group)

  def set_sprites(cls, sprite_cls_name: str) -> None:
    """Sets the active sprite of each EauPlayer instance

    Args:
      sprite_cls_name:
        The name of the class of the sprite to instanciate and assign to each
        EauPlayer instance.
    """
    cls._current_sprite = sprite_cls_name
    sprite_cls = scpup.EauSprite.subclasses.get(sprite_cls_name)
    if sprite_cls:
      for p in cls._instances.values():
        p.set_sprite(sprite_cls(f"{p.pid}.png", masked=True, owner=p.pid))


@final
class EauPlayer(metaclass=EauPlayerMeta):
  """ A player instance

  Attributes:
    pid:
      The player id. It can be one of 'p1', 'p2', 'p3', and 'p4'.
    iid:
      The controller instance id of the controller assigned to this player.
    sprites:
      A named group of the sprites that belong to this player.
  """
  __slots__: tuple = (
    "pid",
    "iid",
    "sprites"
  )

  def __init__(self, player_id: PlayerIdType) -> None:
    """Initializes a player given a player id.

    Args:
      player_id:
        The player id.
    """
    super().__init__()
    self.pid = player_id
    self.iid: int | None = None
    self.sprites = scpup.EauNamedGroup()
    if self.__class__._current_sprite:
      cls = EauSprite.subclasses.get(self.__class__._current_sprite)
      if cls:
        self.set_sprite(cls(f"{self.pid}.png", masked=True, owner=self.pid))

  @property
  def sprite(self) -> scpup.EauSprite | None:
    """Get the active sprite of this player

    Returns:
      EauSprite:
        The active sprite or None if there is no active sprite. Active sprite
        means the sprite that the player is currently using
    """
    group = self.sprites.sprites('active')
    if len(group) > 0:
      return group[0]
    return None

  @property
  def ctrl(self) -> scpup.EauCtrl | None:
    """Get the controller assigned to this player

    Returns:
      EauCtrl:
        The assigned controller. If this player does not have a controller
        assigned this method will return None.
    """
    return None if self.iid is None else scpup.EauCtrl(self.iid)

  @overload
  def assign_ctrl(self, iid: int) -> None:
    """Assign a controller to this player given its instance id

    Args:
      iid:
        The controller instance id.
    """
  @overload
  def assign_ctrl(self, ctrl: scpup.EauCtrl) -> None:
    """Assign a controller to this player

    Args:
      ctrl:
        The controller to assign
    """
  def assign_ctrl(self, iid_or_ctrl: scpup.EauCtrl | int) -> None:  # type: ignore
    if isinstance(iid_or_ctrl, int):
      self.iid = iid_or_ctrl
    else:
      self.iid = iid_or_ctrl.iid

  def unassign_ctrl(self) -> None:
    """Unassign the currently assigned controller of this player"""
    self.iid = None

  @overload
  def handle_joystick_input(self, button_num: int) -> None:
    """Handle a button down event.

    Args:
      button_num:
        The button num corresponding to the button that was pressed.
    """
  @overload
  def handle_joystick_input(self, axis_num: int, value: float) -> None:
    """Handle a button down event.

    Args:
      axis_num:
        The axis num corresponding to the axis that was moved.
      value:
        The amount of movement of the axis
    """
  def handle_joystick_input(self, num: int, value: float | None = None) -> None:  # type: ignore
    ctrl = self.ctrl
    if ctrl:
      action = ctrl.action(num) if value is None else ctrl.action(num, value)
      sprite = self.sprite
      if action and sprite:
        sprite.on_action(action)

  def move(self, area: pygame.Rect | None = None) -> None:
    """Move the player's active sprite

    Args:
      area:
        The area to clamp the sprite in after moving it, if None is passed then
        the sprite won't be clamped. Defaults to None.
    """
    ctrl, sprite = self.ctrl, self.sprite
    if ctrl and sprite:
      sprite.move(ctrl.LS)
      if area:
        sprite.clamp(area)

  def check_collition(self, group: scpup.EauGroup):
    """Check collitions with a group of sprites

    Args:
      group:
        The group of sprites to check the collition with
    """
    sprite = self.sprite
    if sprite:
      collided_sprite = pygame.sprite.spritecollideany(sprite, group, collided=scpup.eau_collide)
      if collided_sprite:
        sprite.on_collition(collided_sprite)
        collided_sprite.on_collition(sprite)

  def set_sprite(self, sprite: scpup.EauSprite):
    """Sets the active sprite

    Args:
      sprite:
        The sprite that's going to be set as the active sprite
    """
    # self.sprites.transfer(self.sprite, 'cached')
    sprt = self.sprite
    if sprt:
      sprt.kill()
    self.sprites.add(sprite, name='active')



from __future__ import annotations

import pygame
import pygame.event
import scpup
from typing import Any, Literal, final, overload


__all__: list[str] = [
  "EauService",
  "EauDisplayService",
  "EauEventService",
  "EAU_EVENT"
]


EAU_EVENT = pygame.event.custom_type()


class EauServiceMeta(type):
  _instances_: dict[str, Any] = {}

  def __call__(cls, *args, **kwds) -> Any:
    if cls.__name__ not in cls._instances_:
      instance = super().__call__(*args, **kwds)
      cls._instances_[cls.__name__] = instance
    return cls._instances_[cls.__name__]


class EauService(metaclass=EauServiceMeta):
  """Base class for scpup services. If you want to use a service you'll need to
  call EauService.get and pass the service name as an argument"""
  @overload
  @classmethod
  def get(cls, service_name: Literal['EauDisplayService']) -> EauDisplayService:
    """Get the display service"""
  @overload
  @classmethod
  def get(cls, service_name: Literal['EauEventService']) -> EauEventService:
    """Get the event service"""
  @classmethod
  def get(cls, service_name: str) -> EauService:
    service = cls._instances_.get(service_name, None)
    if not service:
      raise RuntimeError(f"Service '{service_name}' does not exist or has not been initialized.")
    return service


@final
class EauDisplayService(EauService):
  """A service used for all display related tasks

  Attributes:
    _display:
      The main display of the game
    _background:
      The background of the current view
    _default_bg:
      The default background used in views that don't have a background
    _view:
      The current view that is being rendered
    viewport:
      The main display rect

  Class Attributes:
    icon_path:
      The path of the icon of the game window
    caption:
      The title of the game window
    size:
      the size of the game window in a tuple like this: (width, height). By
      default width is 1200 and height is 800.
  """
  icon_path: str | None = None
  caption: str | None = None
  size: tuple[int, int] = (1200, 800)

  __slots__: tuple = (
    "_display",
    "_background",
    "_view",
    "_default_bg",
    "viewport"
  )

  @classmethod
  def set_window_attrs(
    cls,
    icon_path: str | None = None,
    caption: str | None = None,
    size: tuple[int, int] | None = None
  ) -> None:
    """Set the attributes of the game window. Use this before initializing
    the scpup module. If the parameters are set to None they won't be taken in
    consideration, and by default they are all None.

    Args:
      icon_path:
        The path of the game window icon.
      caption:
        The game window title.
      size:
        The game window size.
    """
    if icon_path:
      cls.icon_path = icon_path
    if caption:
      cls.caption = caption
    if size:
      cls.size = size

  def __init__(self):
    """Initialize the display service"""
    self.viewport = pygame.Rect(0, 0, *self.size)
    self._display: pygame.Surface = pygame.display.set_mode(self.size)
    if self.icon_path:
      img, _ = scpup.load_image(self.icon_path)
      pygame.display.set_icon(img)
    if self.caption:
      pygame.display.set_caption(self.caption)
    pygame.mouse.set_visible(0)
    self._view: scpup.EauView | None = None
    bg = pygame.Surface(self.size)
    bg.fill(pygame.Color(86, 193, 219))
    self._default_bg = bg.convert()
    self._background: pygame.surface.Surface = self._default_bg

  @property
  def view(self) -> scpup.EauView | None:
    """Get the current view"""
    return self._view

  @view.setter
  def view(self, view: scpup.EauView) -> None:
    """Set the current view

    If there was a view already being rendered, this method first clears the
    display and then sets the view. It also sets the players sprites
    """
    if self.view:
      self.view.sprites.clear(self._display, self._background)
    self._view = view
    self._background = view.background or self._default_bg
    self._display.blit(self._background, (0, 0))
    scpup.EauPlayer.set_sprites(view.player_sprite)

  def update_view(self, *args, **kwargs) -> None:
    """Main display update method

    This method calls the `clear` method, then the `update` method, and then the
    `draw` method of the view and the players. This method also checks for
    collitions between the players sprites and the view sprites, and then
    between the players sprites and the other players sprites.

    Args:
      *args, **kwargs:
        Any arguments to pass to the sprites update method
    """
    if self.view:
      self.view.clear(self._display, self._background)
      scpup.EauPlayer.clear(self._display, self._background)
      self.view.update(*args, rect=self.viewport, **kwargs)
      scpup.EauPlayer.update(*args, rect=self.viewport, **kwargs)
      self.view.draw(self._display)
      scpup.EauPlayer.draw(self._display)
      scpup.EauPlayer.check_collitions(self.view.sprites)
    pygame.display.flip()


@final
class EauEventService(EauService):
  """A service used for event related tasks"""
  __slots__ = ()

  def __init__(self) -> None:
    pygame.key.set_repeat()
    pygame.key.stop_text_input()
    pygame.event.set_blocked([
        pygame.MOUSEMOTION,
        pygame.WINDOWLEAVE,
        pygame.WINDOWENTER,
        pygame.WINDOWFOCUSLOST,
        pygame.WINDOWFOCUSGAINED,
        pygame.WINDOWSHOWN,
        pygame.WINDOWCLOSE,
        pygame.ACTIVEEVENT,
        pygame.MOUSEBUTTONDOWN,
        pygame.MOUSEBUTTONUP,
        pygame.VIDEOEXPOSE,
        pygame.VIDEORESIZE,
        pygame.WINDOWEXPOSED,
        pygame.AUDIODEVICEADDED,
        pygame.AUDIODEVICEREMOVED
    ])
    # pygame.event.clear()

  def handle_event(self, event):
    """Main event handler

    This method handles an event. It is intended to be so generic that it can be
    used in scpup as well as in other games, but work still in progress....

    Args:
      event:
        The pygame.event.Event to handle.
    """
    if event.type == pygame.JOYDEVICEADDED:
      j: pygame.joystick.Joystick = pygame.joystick.Joystick(event.device_index)
      scpup.EauCtrl(j)

    elif event.type == pygame.JOYDEVICEREMOVED:
      scpup.EauPlayer.remove_player(event.instance_id)
      scpup.EauCtrl.remove_ctrl(event.instance_id)

    elif event.type in [pygame.JOYAXISMOTION, pygame.JOYBUTTONDOWN]:
      player: scpup.EauPlayer | None = scpup.EauPlayer(event.instance_id)
      if player:
        if event.type == pygame.JOYAXISMOTION:
          player.handle_joystick_input(event.axis, event.value)
        else:
          player.handle_joystick_input(event.button)
      elif event.type == pygame.JOYBUTTONDOWN:
        ctrl = scpup.EauCtrl(event.instance_id)
        if ctrl.action(event.button) == scpup.EauAction.START:
          try:
            new_player: scpup.EauPlayer = next(scpup.EauPlayer)
            new_player.assign_ctrl(ctrl)
          except StopIteration:
            return
      elif event.type == EAU_EVENT:
        ...

  def send(self, event_name: str, **kwargs):
    ev = pygame.event.Event(EAU_EVENT, {
      'name': event_name,
      **kwargs
    })
    pygame.event.post(ev)
